/*
 * bz2libmodule.c -- bzip2-compatible data compression
 * See http://sourceware.cygnus.com/bzip2/
 *
 * This module is base on version 0.9.5d
 *
 * Copyright (C) 2000 Eito Katagiri  All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

/*
 * This module is based on zlibmodule.c. I would like to give the
 * author a credit as well.
 */


#include "Python.h"
#include "bzlib.h"

#define PyInit_bzlib initbzlib

#define BZ_DEFAULT_BLOCKSIZE100K 4 
#define BZ_DEFAULT_VERBOSITY     0
#define BZ_DEFAULT_WORKFACTOR    30

#define BZ_TMPBUFFER_SIZE        512


staticforward PyTypeObject Comptype;


static PyObject *BZlibError;


typedef struct
{
  PyObject_HEAD
  bz_stream bzst;
  PyObject *unused_data;
  int is_initialised;
} compobject;


static char compressobj__doc__[] =
"compressobj() -- Return a compressor object.\n";


static compobject *
newcompobject(type)
  PyTypeObject *type;
{
  compobject *self;

  self = PyObject_NEW(compobject, type);
  if (self == NULL) {
    return (NULL);
  }
  memset((void *) &(self->bzst), '\0', sizeof(bz_stream));
  self->is_initialised = 0;
  self->unused_data = PyString_FromString("");
  
  return (self);
}


static PyObject *
PyBZlib_compress(self, args)
  PyObject *self;
  PyObject *args;
{
  int          blockSize100k = BZ_DEFAULT_BLOCKSIZE100K;
  int          verbosity     = BZ_DEFAULT_VERBOSITY;
  int          workFactor    = BZ_DEFAULT_WORKFACTOR;
  bz_stream    bzst;
  char        *input;
  int          inputLength;
  char        *output;
  unsigned int outputLength;
  char        *tmpOutput;
  char         tmpBuffer[BZ_TMPBUFFER_SIZE];
  char        *tmpBufPtr;
  int          rc;
  int          compressFlag;
  PyObject    *returnVal;


  if (!PyArg_ParseTuple(args, "s#", &input, &inputLength)) {
    return (NULL);
  }

  memset((void *) &bzst, '\0', sizeof(bz_stream));

  rc = bzCompressInit(&bzst, blockSize100k, verbosity, workFactor);
  switch (rc) {
  case BZ_OK:
    break;
  case BZ_PARAM_ERROR:
    PyErr_SetString(BZlibError, "Bad parameters");
    return (NULL);
  case BZ_MEM_ERROR:
    PyErr_SetString(PyExc_MemoryError, "Out of memory while compressing data");
    return (NULL);
  default:
    PyErr_Format(BZlibError, "Unknown error: %i", rc);
    return (NULL);
  }

  memset((void *) tmpBuffer, '\0', sizeof(char) * BZ_TMPBUFFER_SIZE);
  tmpBufPtr      = tmpBuffer;
  bzst.next_in   = input;
  bzst.avail_in  = inputLength;
  bzst.next_out  = tmpBufPtr;
  bzst.avail_out = BZ_TMPBUFFER_SIZE;

  output       = (char *) NULL; 
  outputLength = 0;
  rc = bzCompress(&bzst, BZ_FINISH);
  for (compressFlag = 1; compressFlag;) {
    switch (rc) {
    case BZ_STREAM_END:
      tmpOutput = (char *) realloc(output, sizeof(char) * bzst.total_out);
      if (tmpOutput == (char *) NULL) {
        free(output);
        PyErr_SetString(PyExc_MemoryError,
                        "Out of memory while compressing data");
        return (NULL);
      }
      memcpy((void *) &tmpOutput[outputLength],
             (void *) tmpBuffer,
             sizeof(char) * (bzst.total_out - outputLength));
      output       = tmpOutput;
      outputLength = bzst.total_out;
      compressFlag = 0;
      break;
    case BZ_FINISH_OK:
      tmpOutput = (char *) realloc(output, sizeof(char) * bzst.total_out);
      if (tmpOutput == (char *) NULL) {
        free(output);
        PyErr_SetString(PyExc_MemoryError,
                        "Out of memory while compressing data");
        return (NULL);
      }
      memcpy((void *) &tmpOutput[outputLength],
             (void *) tmpBuffer, 
             sizeof(char) * (bzst.total_out - outputLength));
      output       = tmpOutput;
      outputLength = bzst.total_out;

      memset((void *) tmpBuffer, '\0', sizeof(char) * BZ_TMPBUFFER_SIZE);
      tmpBufPtr      = tmpBuffer;
      bzst.next_out  = tmpBufPtr; 
      bzst.avail_out = BZ_TMPBUFFER_SIZE;
      rc = bzCompress(&bzst, BZ_FINISH);
      break;
    case BZ_SEQUENCE_ERROR:
      free(output);
      PyErr_SetString(BZlibError, "Wrong compress sequence");
      return (NULL);
    default:
      free(output);
      PyErr_Format(BZlibError, "Unknown error: %i", rc);
      return (NULL);
    }
  }

  rc = bzCompressEnd(&bzst);
  switch (rc) {
  case BZ_OK:
    break;
  case BZ_PARAM_ERROR:
    free(output);
    PyErr_SetString(BZlibError, "Wrong parameter");
    return (NULL);
  default:
    free(output);
    PyErr_Format(BZlibError, "Unknown error: %i", rc);
    return (NULL);
  }

  returnVal = PyString_FromStringAndSize(output, outputLength);
  free(output);
  return (returnVal);
}


static PyObject *
PyBZlib_decompress(self, args)
  PyObject *self;
  PyObject *args;
{
  int          verbosity     = BZ_DEFAULT_VERBOSITY;
  bz_stream    bzst;
  char        *input;
  int          inputLength;
  char        *output;
  unsigned int outputLength;
  char        *tmpOutput;
  char         tmpBuffer[BZ_TMPBUFFER_SIZE];
  char        *tmpBufPtr;
  int          rc;
  int          compressFlag;
  PyObject    *returnVal;

  if (!PyArg_ParseTuple(args, "s#", &input, &inputLength)) {
    return (NULL);
  }

  memset((void *) &bzst, '\0', sizeof(bz_stream));

  rc = bzDecompressInit(&bzst, verbosity, 0);
  switch (rc) {
  case BZ_OK:
    break;
  case BZ_PARAM_ERROR:
    PyErr_SetString(BZlibError, "Bad parameters");
    return (NULL);
  case BZ_MEM_ERROR:
    PyErr_SetString(PyExc_MemoryError, "Out of memory while compressing data");
    return (NULL);
  default:
    PyErr_Format(BZlibError, "Unknown error: %i", rc);
    return (NULL);
  }

  memset((void *) tmpBuffer, '\0', sizeof(char) * BZ_TMPBUFFER_SIZE);
  tmpBufPtr      = tmpBuffer;
  bzst.next_in   = input;
  bzst.avail_in  = inputLength;
  bzst.next_out  = tmpBufPtr;
  bzst.avail_out = BZ_TMPBUFFER_SIZE;

  output       = (char *) NULL; 
  outputLength = 0;
  rc = bzDecompress(&bzst);
  for (compressFlag = 1; compressFlag;) {
    switch (rc) {
    case BZ_STREAM_END:
      tmpOutput = (char *) realloc(output, sizeof(char) * outputLength);
      if (tmpOutput == (char *) NULL) {
        free(output);
        PyErr_SetString(PyExc_MemoryError,
                        "Out of memory while compressing data");
        return (NULL);
      }
      memcpy((void *) &tmpOutput[outputLength],
             (void *) tmpBuffer, 
             sizeof(char) * (bzst.total_out - outputLength));
      output       = tmpOutput;
      outputLength = bzst.total_out;
      compressFlag = 0;
      break;
    case BZ_OK:
      tmpOutput = (char *) realloc(output, sizeof(char) * bzst.total_out); 
      if (tmpOutput == (char *) NULL) {
        free(output);
        PyErr_SetString(PyExc_MemoryError,
                        "Out of memory while compressing data");
        return (NULL);
      }
      memcpy((void *) &tmpOutput[outputLength],
             (void *) tmpBuffer, 
             sizeof(char) * (bzst.total_out - outputLength));
      output       = tmpOutput;
      outputLength = bzst.total_out;

      memset((void *) tmpBuffer, '\0', sizeof(char) * BZ_TMPBUFFER_SIZE);
      tmpBufPtr      = tmpBuffer;
      bzst.next_out  = tmpBufPtr; 
      bzst.avail_out = BZ_TMPBUFFER_SIZE;
      rc = bzDecompress(&bzst);
      break;
    case BZ_SEQUENCE_ERROR:
      free(output);
      PyErr_SetString(BZlibError, "Wrong compress sequence");
      return (NULL);
    case BZ_DATA_ERROR:
      free(output);
      PyErr_SetString(BZlibError, "Bad data integrity");
      return (NULL);
    case BZ_DATA_ERROR_MAGIC:
      free(output);
      PyErr_SetString(BZlibError, "Bad magic bytes in data");
      return (NULL);
    default:
      free(output);
      PyErr_Format(BZlibError, "Unknown error: %i", rc);
      return (NULL);
    }
  }

  rc = bzDecompressEnd(&bzst);
  switch (rc) {
  case BZ_OK:
    break;
  case BZ_PARAM_ERROR:
    free(output);
    PyErr_SetString(BZlibError, "Wrong parameter");
    return (NULL);
  default:
    free(output);
    PyErr_Format(BZlibError, "Unknown error: %i", rc);
    return (NULL);
  }

  returnVal = PyString_FromStringAndSize(output, outputLength);
  free(output);
  return (returnVal);
}


static PyObject *
PyBZlib_compressobj(selfptr, args)
  PyObject *selfptr;
  PyObject *args;
{
  compobject *self;
  int rc;

  if (! PyArg_ParseTuple(args, "")) {
    return (NULL);
  }

  self = newcompobject(&Comptype);
  if (self == NULL) {
    return (NULL);
  }

  rc = bzCompressInit(&(self->bzst), BZ_DEFAULT_BLOCKSIZE100K, BZ_DEFAULT_VERBOSITY, BZ_DEFAULT_WORKFACTOR); 
  switch (rc) {
  case BZ_OK:
    self->is_initialised = 1;
    break;
  case BZ_PARAM_ERROR:
    Py_DECREF(self);
    PyErr_SetString(BZlibError, "Bad parameters");
    return (NULL);
  case BZ_MEM_ERROR:
    Py_DECREF(self);
    PyErr_SetString(PyExc_MemoryError, "Out of memory while compressing data");
    return (NULL);
  default:
    Py_DECREF(self);
    PyErr_Format(BZlibError, "Unknown error: %i", rc);
    return (NULL);
  }

  return ((PyObject *) self);
}


static char comp_compress__doc__[] =
"compress(data) -- Return a string containing a compressed version of the data.\
n\n"
"After calling this function, some of the input data may still\n"
"be stored in internal buffers for later processing.\n"
"Call the flush() method to clear these buffers."
;

static PyObject *
PyBZlib_objcompress(self, args)
  compobject *self;
  PyObject *args;
{
  char        *input;
  unsigned int inputLength;
  unsigned int outputLength;
  PyObject    *returnVal;
  int          compressFlag;
  unsigned int startTotalOut;
  int          rc;
  
  if (!PyArg_ParseTuple(args, "s#", &input, &inputLength)) {
    return (NULL);
  }
  
  outputLength = BZ_TMPBUFFER_SIZE;
  returnVal = PyString_FromStringAndSize((char *) NULL, outputLength); 
  if (returnVal == NULL) {
    PyErr_SetString(PyExc_MemoryError,
                    "Can't allocate memory to compress data");
    return (NULL);
  }
  startTotalOut = self->bzst.total_out;
  self->bzst.next_in   = input;
  self->bzst.avail_in  = inputLength;
  self->bzst.next_out  = (char *) PyString_AsString(returnVal);
  self->bzst.avail_out = BZ_TMPBUFFER_SIZE;

  rc = bzCompress(&(self->bzst), BZ_RUN);
  for (compressFlag = 1; compressFlag;) {
    switch (rc) {
    case BZ_RUN_OK:
      if (self->bzst.avail_in <= 0) {
        compressFlag = 0;
        break;
      }
      if (self->bzst.avail_out <= 0) {
        if (_PyString_Resize(&returnVal, outputLength << 1) == -1) { 
          PyErr_SetString(PyExc_MemoryError,
                          "Can't allocate memory to compress data");
          return (NULL);
        }
        self->bzst.next_out = (char *)PyString_AsString(returnVal) + outputLength;
        self->bzst.avail_out = outputLength;
        outputLength <<= 1;
      } else {
        self->bzst.next_out = self->bzst.next_out + self->bzst.total_out;
      }
      break;
    case BZ_SEQUENCE_ERROR:
      PyErr_SetString(BZlibError, "Wrong compress sequence");
      Py_DECREF(returnVal);
      return (NULL);
    default:
      PyErr_Format(BZlibError, "Unknown error: %i", rc);
      Py_DECREF(returnVal);
      return (NULL);
    }
  }
  _PyString_Resize(&returnVal, self->bzst.total_out - startTotalOut); 
  return (returnVal);
}


static char comp_flush__doc__[] =
"flush() -- Return a string containing any remaining compressed data.\n";

static PyObject *
PyBZlib_flush(self, args)
  compobject *self;
  PyObject *args;
{
  unsigned int outputLength;
  PyObject *returnVal;
  int compressFlag;
  int rc;
  unsigned int startTotalOut;
  
  if (!PyArg_ParseTuple(args, "")) {
    return (NULL);
  }

  outputLength = BZ_TMPBUFFER_SIZE;
  returnVal = PyString_FromStringAndSize((char *) NULL, outputLength); 
  if (returnVal == NULL) {
    PyErr_SetString(PyExc_MemoryError,
                    "Can't allocate memory to compress data");
    return (NULL);
  }
  startTotalOut = self->bzst.total_out;
  self->bzst.next_in   = (char *) NULL;
  self->bzst.avail_in  = 0;
  self->bzst.next_out  = (char *) PyString_AsString(returnVal);
  self->bzst.avail_out = BZ_TMPBUFFER_SIZE;

  rc = bzCompress(&(self->bzst), BZ_FINISH);
  for (compressFlag = 1; compressFlag;) {
    switch (rc) {
    case BZ_STREAM_END:
      compressFlag = 0;
      break;
    case BZ_FINISH_OK:
      if (self->bzst.avail_out <= 0) {
        if (_PyString_Resize(&returnVal, outputLength << 1) == -1) { 
          PyErr_SetString(PyExc_MemoryError,
                          "Can't allocate memory to compress data");
          return (NULL);
        }
        self->bzst.next_out = (char *)PyString_AsString(returnVal) + outputLength;
        self->bzst.avail_out = outputLength;
        outputLength <<= 1;
      } else {
        self->bzst.next_out = self->bzst.next_out + self->bzst.total_out;
      }
      break;
    case BZ_SEQUENCE_ERROR:
      PyErr_SetString(BZlibError, "Wrong compress sequence");
      Py_DECREF(returnVal);
      return (NULL);
    default:
      PyErr_Format(BZlibError, "Unknown error: %i", rc);
      Py_DECREF(returnVal);
      return (NULL);
    }
  }
  _PyString_Resize(&returnVal, self->bzst.total_out - startTotalOut); 
  return (returnVal);
}


static void
Comp_dealloc(self)
  compobject *self;
{
  Py_XDECREF(self->unused_data);
  PyMem_DEL(self);
}


static PyMethodDef comp_methods[] = 
{
  {"compress", (binaryfunc) PyBZlib_objcompress, 1, comp_compress__doc__},
  {"flush", (binaryfunc) PyBZlib_flush, 1, comp_flush__doc__},
  {NULL, NULL}
};


static PyObject *
Comp_getattr(self, name)
  compobject *self;
  char *name;
{
  return (Py_FindMethod(comp_methods, (PyObject *) self, name));
}

static PyMethodDef bzlib_methods[] =
{
  {"compress", (PyCFunction) PyBZlib_compress, 1, NULL},
  {"decompress", (PyCFunction) PyBZlib_decompress, 1, NULL},
  {"compressobj", (PyCFunction) PyBZlib_compressobj, 1, NULL}, 
  {NULL, NULL}
};


statichere PyTypeObject Comptype = {
  PyObject_HEAD_INIT(0)
  0,
  "Compress",
  sizeof(compobject),
  0,
  (destructor) Comp_dealloc,      /*tp_dealloc*/
  0,                              /*tp_print*/
  (getattrfunc) Comp_getattr,     /*tp_getattr*/
  0,                              /*tp_setattr*/
  0,                              /*tp_compare*/
  0,                              /*tp_repr*/
  0,                              /*tp_as_number*/
  0,                              /*tp_as_sequence*/
  0,                              /*tp_as_mapping*/
};


static char bzlib_module_documentation[] =
  "compress(string) - Compress a string\n"
  "decompress(string) - Decompress a string\n";


DL_EXPORT(void)
PyInit_bzlib()
{
  PyObject *mod;
  PyObject *dic;
  PyObject *ver;

  Comptype.ob_type = &PyType_Type;

  mod = Py_InitModule4("bzlib", bzlib_methods,
                      bzlib_module_documentation, (PyObject *) NULL,
                      PYTHON_API_VERSION);
  dic = PyModule_GetDict(mod);

  BZlibError = PyErr_NewException("bzlib.error", NULL, NULL);
  PyDict_SetItemString(dic, "error", BZlibError);

  ver = PyString_FromString("0.95.5d");
  PyDict_SetItemString(dic, "BZLIB_VERSION", ver);
  Py_DECREF(ver);
}
